#!/bin/env python3

import argparse
import datetime
import os

import numpy as np
import open3d as o3d

voxel_size = 0.02
max_correspondence_distance_coarse = voxel_size * 15
max_correspondence_distance_fine = voxel_size * 1.5
threshold = 0.02

def load_point_clouds(path, downsample):
    pcds = []
    all_files = os.listdir(path)

    for file in all_files:
        file_path = os.path.join(path, file)
        if os.path.isfile(file_path) and file.endswith('.ply'):
            pcd = o3d.io.read_point_cloud(file_path)
            
            if downsample:
                pcd = pcd.voxel_down_sample(voxel_size=downsample)
            pcds.append(pcd)

    pcds = pcds[5:-5] # the first and last 3 seconds of the frames have hand, remove
    return pcds


def pairwise_registration(source, target):
    print("Apply point-to-plane ICP")
    icp_coarse = o3d.pipelines.registration.registration_icp(
        source, target, max_correspondence_distance_coarse, np.identity(4),
        o3d.pipelines.registration.TransformationEstimationPointToPlane())
    icp_fine = o3d.pipelines.registration.registration_icp(
        source, target, max_correspondence_distance_fine,
        icp_coarse.transformation,
        o3d.pipelines.registration.TransformationEstimationPointToPlane())
    transformation_icp = icp_fine.transformation
    information_icp = o3d.pipelines.registration.get_information_matrix_from_point_clouds(
        source, target, max_correspondence_distance_fine,
        icp_fine.transformation)
    return transformation_icp, information_icp


def full_registration(pcds):
    pose_graph = o3d.pipelines.registration.PoseGraph()
    odometry = np.identity(4)
    pose_graph.nodes.append(o3d.pipelines.registration.PoseGraphNode(odometry))
    n_pcds = len(pcds)
    for source_id in range(n_pcds):
        for target_id in range(source_id + 1, n_pcds):
            transformation_icp, information_icp = pairwise_registration(
                pcds[source_id], pcds[target_id])
            print("Build o3d.pipelines.registration.PoseGraph")
            if target_id == source_id + 1:  # odometry case
                odometry = np.dot(transformation_icp, odometry)
                pose_graph.nodes.append(
                    o3d.pipelines.registration.PoseGraphNode(
                        np.linalg.inv(odometry)))
                pose_graph.edges.append(
                    o3d.pipelines.registration.PoseGraphEdge(source_id,
                                                             target_id,
                                                             transformation_icp,
                                                             information_icp,
                                                             uncertain=False))
            else:  # loop closure case
                pose_graph.edges.append(
                    o3d.pipelines.registration.PoseGraphEdge(source_id,
                                                             target_id,
                                                             transformation_icp,
                                                             information_icp,
                                                             uncertain=True))
    return pose_graph


def multi_way_align(pcds):
    # construct graph
    pose_graph = full_registration(pcds)

    print("Optimizing PoseGraph ...")
    option = o3d.pipelines.registration.GlobalOptimizationOption(
        max_correspondence_distance=max_correspondence_distance_fine,
        edge_prune_threshold=0.25,
        reference_node=0)

    o3d.pipelines.registration.global_optimization(
        pose_graph,
        o3d.pipelines.registration.GlobalOptimizationLevenbergMarquardt(),
        o3d.pipelines.registration.GlobalOptimizationConvergenceCriteria(),
        option)
    
    for point_id in range(len(pcds)):
        print(pose_graph.nodes[point_id].pose)
        pcds[point_id].transform(pose_graph.nodes[point_id].pose)


def icp(source, target):
    # print("Initial alignment")
    # evaluation = o3d.pipelines.registration.evaluate_registration(
    #     source, target, threshold, np.eye(4))
    # print(evaluation)

    reg_p2p = o3d.pipelines.registration.registration_icp(
    source, target, threshold, np.eye(4),
    o3d.pipelines.registration.TransformationEstimationPointToPoint())
    source.transform(reg_p2p.transformation)

    # print("After icp")
    # print(reg_p2p)

def align(pcds):
    for i in range(len(pcds) - 1):
        icp(pcds[i+1], pcds[i])

def create_mesh(pcds):
    meshes = []
    with o3d.utility.VerbosityContextManager(o3d.utility.VerbosityLevel.Info) as cm:
        for pcd in pcds:
            mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_alpha_shape(pcd, 0.3)
            meshes.append(mesh)
    return meshes

def count_total_num_points(pcds):
    total = 0
    for pcd in pcds:
        total += len(pcd.points)
    return total

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Merge Point Cloud')
    parser.add_argument('path')
    parser.add_argument('--align', action='store_true')
    parser.add_argument('--mesh', action='store_true')
    parser.add_argument('--downsample', type=float, default=voxel_size)
    args = parser.parse_args()

    pcds = load_point_clouds(args.path, args.downsample)

    if args.align:
        start_time = datetime.datetime.now()
        align(pcds)
        end_time = datetime.datetime.now()
        time_diff = (end_time - start_time)
        print(f"ICP execution {time_diff.total_seconds()} sec")


    if args.mesh:
        meshes = create_mesh(pcds)
        o3d.visualization.draw_geometries(meshes)
    else:
        num_points = count_total_num_points(pcds)
        print(f"we have {len(pcds)} point clouds and {num_points} total points")
        o3d.visualization.draw_geometries(pcds)
